"""Module of transformers.

A transformer transforms time series to extract useful information.

"""
from .._transformer_base import (
    _NonTrainableMultivariateTransformer,
    _NonTrainableUnivariateTransformer,
    _TrainableMultivariateTransformer,
    _TrainableUnivariateTransformer,
)
from .._utils import _get_all_subclasses_from_superclass
from ._transformer_1d import (
    ClassicSeasonalDecomposition,
    CustomizedTransformer1D,
    DoubleRollingAggregate,
    Retrospect,
    RollingAggregate,
    StandardScale,
)
from ._transformer_hd import (
    CustomizedTransformerHD,
    PcaProjection,
    PcaReconstruction,
    PcaReconstructionError,
    RegressionResidual,
    SumAll,
)


def print_all_models() -> None:
    """
    Print description of every model in this module.
    """
    model_desc = _get_all_subclasses_from_superclass(
        _NonTrainableUnivariateTransformer
    )
    model_desc.update(
        _get_all_subclasses_from_superclass(
            _NonTrainableMultivariateTransformer
        )
    )
    model_desc.update(
        _get_all_subclasses_from_superclass(_TrainableUnivariateTransformer)
    )
    model_desc.update(
        _get_all_subclasses_from_superclass(_TrainableMultivariateTransformer)
    )
    for key, value in model_desc.items():
        print("-" * 80)
        print(key)
        print(value)


__all__ = [
    "RollingAggregate",
    "DoubleRollingAggregate",
    "ClassicSeasonalDecomposition",
    "Retrospect",
    "StandardScale",
    "CustomizedTransformer1D",
    "RegressionResidual",
    "PcaProjection",
    "PcaReconstruction",
    "PcaReconstructionError",
    "SumAll",
    "CustomizedTransformerHD",
    "print_all_models",
]



"""
Anomaly Detection Toolkit
=========================

Anomaly Detection Toolkit (ADTK) is a Python package for unsupervised /
rule-based time series anomaly detection.

As the nature of anomaly varies over different cases, a model may not work
universally for all anomaly detection problems. Choosing and combining
detection algorithms (detectors), feature engineering methods (transformers),
and ensemble methods (aggregators) properly is the key to build an effective
anomaly detection model.

This package offers a set of common detectors, transformers and aggregators
with unified APIs, as well as pipe classes that connect them together into
models. It also provides some functions to process and visualize time series
and anomaly events.

See https://adtk.readthedocs.io for complete documentation.

"""

__version__ = "0.6.2"



"""Module is for data (time series and anomaly list) processing.
"""

from typing import Dict, List, Optional, Tuple, Union, overload

import numpy as np
import pandas as pd


def validate_series(
    ts: Union[pd.Series, pd.DataFrame],
    check_freq: bool = True,
    check_categorical: bool = False,
) -> Union[pd.Series, pd.DataFrame]:
    """Validate time series.

    This functoin will check some common critical issues of time series that
    may cause problems if anomaly detection is performed without fixing them.
    The function will automatically fix some of them and raise errors for the
    others.

    Issues will be checked and automatically fixed include:

    - Time index is not monotonically increasing;
    - Time index contains duplicated time stamps (fix by keeping first values);
    - (optional) Time index attribute `freq` is missed while the index follows
      a frequency;
    - (optional) Time series include categorical (non-binary) label columns
      (to fix by converting categorical labels into binary indicators).

    Issues will be checked and raise error include:

    - Wrong type of time series object (must be pandas Series or DataFrame);
    - Wrong type of time index object (must be pandas DatetimeIndex).

    Parameters
    ----------
    ts: pandas Series or DataFrame
        Time series to be validated.

    check_freq: bool, optional
        Whether to check time index attribute `freq` is missed. Default: True.

    check_categorical: bool, optional
        Whether to check time series include categorical (non-binary) label
        columns. Default: False.

    Returns
    -------
    pandas Series or DataFrame
        Validated time series.

    """

    ts = ts.copy()

    # check input type
    if not isinstance(ts, (pd.Series, pd.DataFrame)):
        raise TypeError("Input is not a pandas Series or DataFrame object")

    # check index type
    if not isinstance(ts.index, pd.DatetimeIndex):
        raise TypeError(
            "Index of time series must be a pandas DatetimeIndex object."
        )

    # check duplicated
    if any(ts.index.duplicated(keep="first")):
        ts = ts[ts.index.duplicated(keep="first") == False]

    # check sorted
    if not ts.index.is_monotonic_increasing:
        ts.sort_index(inplace=True)

    # check time step frequency
    if check_freq:
        if (ts.index.freq is None) and (ts.index.inferred_freq is not None):
            ts = ts.asfreq(ts.index.inferred_freq)

    # convert categorical labels into binary indicators
    if check_categorical:
        if isinstance(ts, pd.DataFrame):
            ts = pd.get_dummies(ts)
        if isinstance(ts, pd.Series):
            seriesName = ts.name
            ts = pd.get_dummies(
                ts.to_frame(),
                prefix="" if seriesName is None else seriesName,
                prefix_sep="" if seriesName is None else "_",
            )
            if len(ts.columns) == 1:
                ts = ts[ts.columns[0]]
                ts.name = seriesName

    return ts


def validate_events(
    event_list: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    point_as_interval: bool = False,
) -> List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]:
    """Validate event list.

    This function will check and fix some common issues in an event list (a
    list of time windows), including invalid time window, overlapped time
    windows, unsorted events, etc.

    Parameters
    ----------
    event_list: list
        A list of events, where an event is a pandas Timestamp if it is
        instantaneous or a 2-tuple of pandas Timestamps if it is a closed time
        interval.

    point_as_interval: bool, optional
        Whether to return all instantaneous event as a close interval with
        identicial start point and end point. Default: False.

    Returns
    -------
    list:
        A validated list of events.

    """
    if not isinstance(event_list, list):
        raise TypeError("Argument `event_list` must be a list.")
    for event in event_list:
        if not (
            isinstance(event, pd.Timestamp)
            or (
                isinstance(event, tuple)
                and (len(event) == 2)
                and all([isinstance(event[i], pd.Timestamp) for i in [0, 1]])
            )
        ):
            raise TypeError(
                "Every event in the list must be a pandas Timestamp, "
                "or a 2-tuple of Timestamps."
            )

    time_window_ends = []  # type: List[pd.Timestamp]
    time_window_type = []  # type: List[int]
    for time_window in event_list:
        if isinstance(time_window, tuple):
            if time_window[0] <= time_window[1]:
                time_window_ends.append(time_window[0])
                time_window_type.append(+1)
                time_window_ends.append(time_window[1])
                time_window_type.append(-1)
        else:
            time_window_ends.append(time_window)
            time_window_type.append(+1)
            time_window_ends.append(time_window)
            time_window_type.append(-1)
    time_window_end_series = pd.Series(
        time_window_type, index=pd.DatetimeIndex(time_window_ends), dtype=int
    )  # type: pd.Series
    time_window_end_series.sort_index(kind="mergesort", inplace=True)
    time_window_end_series = time_window_end_series.cumsum()
    status = 0
    merged_event_list = (
        []
    )  # type: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    for t, v in time_window_end_series.iteritems():  # type: pd.Timestamp, int
        if (status == 0) and (v > 0):
            start = t  # type: pd.Timestamp
            status = 1
        if (status == 1) and (v <= 0):
            end = t  # type: pd.Timestamp
            merged_event_list.append([start, end])
            status = 0
    for i in range(1, len(merged_event_list)):
        this_start = merged_event_list[i][0]  # type: pd.Timestamp
        this_end = merged_event_list[i][1]  # type: pd.Timestamp
        last_start = merged_event_list[i - 1][0]  # type: pd.Timestamp
        last_end = merged_event_list[i - 1][1]  # type: pd.Timestamp
        if last_end + pd.Timedelta("1ns") >= this_start:
            merged_event_list[i] = [last_start, this_end]
            merged_event_list[i - 1] = None
    merged_event_list = [
        w[0] if (w[0] == w[1] and not point_as_interval) else tuple(w)
        for w in merged_event_list
        if w is not None
    ]
    return merged_event_list


@overload
def to_events(
    labels: pd.Series,
    freq_as_period: bool = True,
    merge_consecutive: Optional[bool] = None,
) -> List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]:
    ...


@overload
def to_events(  # type: ignore
    labels: pd.DataFrame,
    freq_as_period: bool = True,
    merge_consecutive: Optional[bool] = None,
) -> Dict[str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]]:
    ...


def to_events(
    labels: Union[pd.Series, pd.DataFrame],
    freq_as_period: bool = True,
    merge_consecutive: Optional[bool] = None,
) -> Union[
    List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    Dict[str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]],
]:
    """Convert binary label series to event list.

    Parameters
    ----------
    labels: pandas Series or DataFrame
        Binary series of anomaly labels. If a DataFrame, each column is
        regarded as a type of anomaly independently.

    freq_as_period: bool, optional
        Whether to regard time index with regular frequency (i.e. attribute
        `freq` of time index is not None) as time intervals.

        For example, DatetimeIndex(['2017-01-01', '2017-01-02', '2017-01-03',
        '2017-01-04', '2017-01-05'], dtype='datetime64[ns]', freq='D') has
        daily frequency. If freq_as_period=True, each time point in the index
        represents that day (24 hours). Otherwsie, each time point represents
        the instantaneous time instance of 00:00:00 on that day.

        Default: True.

    merge_consecutive: bool, optional
        Whether to merge consecutive events into a single time window. If not
        specified, it is on automatically if the input time index has a regular
        frequency and freq_as_period=True, and it is off otherwise. Default:
        None.

    Returns
    -------
    list or dict
        - If input is a Series, output is a list of events where an event is a
          pandas Timestamp if it is instantaneous or a 2-tuple of pandas
          Timestamps if it is a closed time interval.
        - If input is a DataFrame, every column is treated as an independent
          binary series, and output is a dict where keys are column names and
          values are event lists.

    """

    if isinstance(labels, pd.Series):
        labels = validate_series(
            labels, check_freq=False, check_categorical=False
        )
        labels = labels == 1

        if merge_consecutive is None:
            if freq_as_period and (labels.index.freq is not None):
                merge_consecutive = True
            else:
                merge_consecutive = False

        if not merge_consecutive:
            if freq_as_period and (labels.index.freq is not None):
                period_end = pd.date_range(
                    start=labels.index[1],
                    periods=len(labels.index),
                    freq=labels.index.freq,
                ) - pd.Timedelta(
                    "1ns"
                )  # type: pd.DatetimeIndex
                return [
                    (start, end) if start != end else start
                    for start, end in zip(
                        list(labels.index[labels]), list(period_end[labels])
                    )
                ]
            else:
                return list(labels.index[labels])
        else:
            labels_values = labels.values.astype(int).reshape(
                -1, 1
            )  # type: np.ndarray
            mydiff = np.vstack(
                [
                    labels_values[0, :] - 0,
                    np.diff(labels_values, axis=0),
                    0 - labels_values[-1, :],
                ]
            )  # type: np.ndarray
            starts = np.argwhere(mydiff == 1)  # type: np.ndarray
            ends = np.argwhere(mydiff == -1)  # type: np.ndarray
            if freq_as_period and (labels.index.freq is not None):
                period_end = pd.date_range(
                    start=labels.index[1],
                    periods=len(labels.index),
                    freq=labels.index.freq,
                ) - pd.Timedelta("1ns")
                return [
                    (labels.index[start], period_end[end - 1])
                    if labels.index[start] != period_end[end - 1]
                    else labels.index[start]
                    for start, end in zip(starts[:, 0], ends[:, 0])
                ]
            else:
                return [
                    (labels.index[start], labels.index[end - 1])
                    if start != end - 1
                    else labels.index[start]
                    for start, end in zip(starts[:, 0], ends[:, 0])
                ]
    else:
        if labels.columns.duplicated().any():
            raise ValueError("Input DataFrame must have unique column names.")
        return {
            col: to_events(labels[col], freq_as_period, merge_consecutive)
            for col in labels.columns
        }


@overload
def to_labels(
    lists: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    time_index: pd.DatetimeIndex,
    freq_as_period: bool = True,
) -> pd.Series:
    ...


@overload
def to_labels(
    lists: Dict[
        str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ],
    time_index: pd.DatetimeIndex,
    freq_as_period: bool = True,
) -> pd.DataFrame:
    ...


def to_labels(
    lists: Union[
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ],
    time_index: pd.DatetimeIndex,
    freq_as_period: bool = True,
) -> Union[pd.Series, pd.DataFrame]:
    """Convert event list to binary series along a time index.

    Parameters
    ----------
    lists: list or dict
        A list of events, or a dict of lists of events.

        - If list, a list of events where an event is a pandas Timestamp if it
          is instantaneous or a 2-tuple of pandas Timestamps if it is a closed
          time interval.
        - If dict, each key-value pair represents an independent list of
          events.

    time_index: pandas DatatimeIndex
        Time index to build the label series.

    freq_as_period: bool, optional
        Whether to regard time index with regular frequency (i.e. attribute
        `freq` of time index is not None) as time intervals.

        For example, DatetimeIndex(['2017-01-01', '2017-01-02', '2017-01-03',
        '2017-01-04', '2017-01-05'], dtype='datetime64[ns]', freq='D') has
        daily frequency. If freq_as_period=True, each time piont represents
        that day, and that day will be marked positive if an event in the event
        list overlaps with the period of that day (24 hours). Otherwsie, each
        time point represents the instantaneous time instance of 00:00:00 on
        that day, and that time point will be marked positive if an event in
        the event list covers it.

        Default: True.

    Returns
    -------
    pandas Series or DataFrame
        Series of binary labels.

        - If input is asingle list, the output is a Series.
        - If input is a dict of lists, the output is a DataFrame where each
          column corresponds a list in the dict.

    """

    if not isinstance(time_index, pd.DatetimeIndex):
        raise TypeError("Time index must be a pandas DatetimeIndex object.")

    if not time_index.is_monotonic_increasing:
        raise ValueError("Time index must be monotoic increasing.")

    if isinstance(lists, list):
        labels = pd.Series(False, index=time_index)  # type: pd.Series
        lists = validate_events(lists)
        if freq_as_period and (time_index.freq is not None):
            period_end = pd.date_range(
                start=time_index[1],
                periods=len(time_index),
                freq=time_index.freq,
            ) - pd.Timedelta(
                "1ns"
            )  # type: pd.DatetimeIndex
            for event in lists:
                isOverlapped = pd.Series(
                    index=time_index, dtype=bool
                )  # type: pd.Series
                if isinstance(event, tuple):
                    isOverlapped = (time_index <= event[1]) & (
                        period_end >= event[0]
                    )
                else:
                    isOverlapped = (time_index <= event) & (
                        period_end >= event
                    )
                labels.loc[isOverlapped] = True
        else:
            for event in lists:
                if isinstance(event, tuple):
                    labels.loc[
                        (labels.index >= event[0]) & (labels.index <= event[1])
                    ] = True
                else:
                    labels.loc[labels.index == event] = True
        return labels
    elif isinstance(lists, dict):
        labels_df = pd.DataFrame(
            False, index=time_index, columns=lists.keys()
        )  # pd.DataFrame
        for col, key in zip(labels_df.columns, lists.keys()):
            labels_df[col] = to_labels(lists[key], time_index, freq_as_period)
        return labels_df
    else:
        raise TypeError("Argument `lists` must be a list or a dict.")


@overload
def expand_events(
    lists: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    left_expand: Union[pd.Timedelta, str, int] = 0,
    right_expand: Union[pd.Timedelta, str, int] = 0,
    freq_as_period: bool = True,
) -> List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]:
    ...


@overload
def expand_events(
    lists: Dict[
        str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ],
    left_expand: Union[pd.Timedelta, str, int] = 0,
    right_expand: Union[pd.Timedelta, str, int] = 0,
    freq_as_period: bool = True,
) -> Dict[str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]]:
    ...


@overload
def expand_events(
    lists: pd.Series,
    left_expand: Union[pd.Timedelta, str, int] = 0,
    right_expand: Union[pd.Timedelta, str, int] = 0,
    freq_as_period: bool = True,
) -> pd.Series:
    ...


@overload
def expand_events(  # type:ignore
    lists: pd.DataFrame,
    left_expand: Union[pd.Timedelta, str, int] = 0,
    right_expand: Union[pd.Timedelta, str, int] = 0,
    freq_as_period: bool = True,
) -> pd.DataFrame:
    ...


def expand_events(  # type:ignore
    events: Union[
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
        pd.Series,
        pd.DataFrame,
    ],
    left_expand: Union[pd.Timedelta, str, int] = 0,
    right_expand: Union[pd.Timedelta, str, int] = 0,
    freq_as_period: bool = True,
) -> Union[
    List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    Dict[str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]],
    pd.Series,
    pd.DataFrame,
]:
    """Expand duration of events.

    Parameters
    ----------
    events: list, dict, pandas Series, or pandas DataFrame
        Events to be expanded.

        - If list, a list of events where an event is a pandas Timestamp if it
          is instantaneous or a 2-tuple of pandas Timestamps if it is a closed
          time interval.
        - If dict, each key-value pair represents an independent list of
          events.
        - If pandas Series, it is binary where 1 represents events cover this
          time point.
        - If pandas DataFrame, each column is treated as an independent Series.

    left_expand: pandas Timedelta, str, or int, optional
        Time range to expand backward.

        - If str, it must be able to be converted into a pandas Timedelta
          object.
        - If int, it must be in nanosecond.

        Default: 0.

    right_expand: pandas Timedelta, str, or int, optional
        Time range to expand forward.

        - If str, it must be able to be converted into a pandas Timedelta
          object.
        - If int, it must be in nanosecond.

        Default: 0.

    freq_as_period: bool, optional
        Whether to regard time index with regular frequency (i.e. attribute
        `freq` of time index is not None) as time intervals. Only used when
        input events is pandas Series or DataFrame.

        For example, DatetimeIndex(['2017-01-01', '2017-01-02', '2017-01-03',
        '2017-01-04', '2017-01-05'], dtype='datetime64[ns]', freq='D') has
        daily frequency. If freq_as_period=True, each time point in the index
        represents that day (24 hours). Otherwsie, each time point represents
        the instantaneous time instance of 00:00:00 on that day.

        Default: True.

    Returns
    -------
    list, dict, pandas Series, or pandas DataFrame
        Expanded events.

    """

    if not isinstance(left_expand, pd.Timedelta):
        left_expand = pd.Timedelta(left_expand)
    if not isinstance(right_expand, pd.Timedelta):
        right_expand = pd.Timedelta(right_expand)

    if isinstance(events, pd.Series):
        labels = validate_series(events)  # type: pd.Series
        lists = to_events(
            labels, freq_as_period=freq_as_period
        )  # type:List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        expanded_lists = expand_events(  # type:ignore
            events=lists, left_expand=left_expand, right_expand=right_expand
        )  # type:List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        expanded_labels = to_labels(
            lists=expanded_lists,
            time_index=labels.index,
            freq_as_period=freq_as_period,
        )  # type: pd.Series
        expanded_labels.loc[
            (expanded_labels == False) & (labels.isna())
        ] = float("nan")
        expanded_labels = expanded_labels.rename(labels.name)
        expanded_labels.index = labels.index
        return expanded_labels
    elif isinstance(events, pd.DataFrame):
        expanded_df = pd.concat(
            [
                expand_events(
                    s,
                    left_expand=left_expand,
                    right_expand=right_expand,
                    freq_as_period=freq_as_period,
                )
                for _, s in events.iteritems()
            ],
            axis=1,
        )  # type: pd.DataFrame
        return expanded_df
    elif isinstance(events, list):
        expanded_list = (
            []
        )  # type: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        for ano in events:
            if isinstance(ano, tuple):
                expanded_list.append(
                    (ano[0] - left_expand, ano[1] + right_expand)
                )
            else:
                expanded_list.append((ano - left_expand, ano + right_expand))
        expanded_list = validate_events(expanded_list)
        return expanded_list
    elif isinstance(events, dict):
        return {
            key: expand_events(value, left_expand, right_expand)
            for key, value in events.items()
        }
    else:
        raise TypeError("Arugment `events` must be a list or a dict of lists.")


def split_train_test(
    ts: Union[pd.Series, pd.DataFrame],
    mode: int = 1,
    n_splits: int = 1,
    train_ratio: float = 0.7,
) -> List[
    Tuple[Union[pd.Series, pd.DataFrame], Union[pd.Series, pd.DataFrame]]
]:
    """Split time series into training and testing set for cross validation.

    Parameters
    ----------
    ts: pandas Series or DataFrame
        Time series to process.

    mode: int, optional
        The split mode to use. Choose from 1, 2, 3 and 4.

        1. Divide time series into n_splits folds of equal length, split
           each fold into training and testing based on train_ratio.

        2. Create n_splits folds, where each fold starts at t_0 and ends
           at t_(n/n_splits), where n goes from 0 to n_splits and the
           first train_ratio of the fold is for training.

        3. Create n_splits folds, where each fold starts at t_0. Each fold
           has len(ts)/(1 + n_splits) test points at the end. Each fold is
           n * len(ts)/(1 + n_splits) long, where n ranges from 1 to
           n_splits.

        4. Create n_splits folds, where each fold starts at t_0. Each fold
           has n * len(ts)/(1 + n_splits) training points at the beginning
           of the time series, where n ranges from 1 to n_splits and the
           remaining points are testing points.

        Default: 1.

    n_splits: int, optional
        Number of splits. Default: 1.

    train_ratio: float, optional
        Ratio between length of training series and each fold, only used by
        mode 1 and 2. Default: 0.7.

    Returns
    -------
        list of 2-tuples (train, test)
            Splitted training and testing series.

    Examples
    --------
        In the following description of the four modes, 1s represent positions
        assigned to training, 2s represent those assigned to testing, 0s are
        those not assigned.

        For a time series with length 40, if `n_splits=4`, `train_ratio=0.7`,

        - If `mode=1`:

            1111111222000000000000000000000000000000
            0000000000111111122200000000000000000000
            0000000000000000000011111112220000000000
            0000000000000000000000000000001111111222

        - If `mode=2`:

            1111111222000000000000000000000000000000
            1111111111111122222200000000000000000000
            1111111111111111111112222222220000000000
            1111111111111111111111111111222222222222

        - If `mode=3`:

            1111111122222222000000000000000000000000
            1111111111111111222222220000000000000000
            1111111111111111111111112222222200000000
            1111111111111111111111111111111122222222

        - If `mode=4`:

            1111111122222222222222222222222222222222
            1111111111111111222222222222222222222222
            1111111111111111111111112222222222222222
            1111111111111111111111111111111122222222

    """

    if not isinstance(ts, (pd.DataFrame, pd.Series)):
        raise ValueError("Argument `ts` must be a pandas Series or DataFrame.")

    splits = []
    if mode == 1:
        fold_len = round(len(ts) / n_splits)
        fold_pos = 0
        for _ in range(n_splits - 1):
            splits.append(
                (
                    ts.iloc[
                        fold_pos : (fold_pos + round(fold_len * train_ratio))
                    ],
                    ts.iloc[
                        (fold_pos + round(fold_len * train_ratio)) : (
                            fold_pos + fold_len
                        )
                    ],
                )
            )
            fold_pos = fold_pos + fold_len
        splits.append(
            (
                ts.iloc[
                    fold_pos : (
                        fold_pos + round((len(ts) - fold_pos) * train_ratio)
                    )
                ],
                ts.iloc[
                    (fold_pos + round((len(ts) - fold_pos) * train_ratio)) :
                ],
            )
        )
    elif mode == 2:
        for k_fold in range(n_splits - 1):
            fold_len = round(len(ts) / n_splits) * (k_fold + 1)
            splits.append(
                (
                    ts.iloc[: round(fold_len * train_ratio)],
                    ts.iloc[round(fold_len * train_ratio) : fold_len],
                )
            )
        splits.append(
            (
                ts.iloc[: round(len(ts) * train_ratio)],
                ts.iloc[round(len(ts) * train_ratio) : len(ts)],
            )
        )
    elif mode == 3:
        fold_len = round(len(ts) / (n_splits + 1))
        for k_fold in range(n_splits - 1):
            splits.append(
                (
                    ts.iloc[: ((k_fold + 1) * fold_len)],
                    ts.iloc[
                        ((k_fold + 1) * fold_len) : ((k_fold + 2) * fold_len)
                    ],
                )
            )
        splits.append(
            (
                ts.iloc[: (n_splits * fold_len)],
                ts.iloc[(n_splits * fold_len) :],
            )
        )
    elif mode == 4:
        fold_len = round(len(ts) / (n_splits + 1))
        for k_fold in range(n_splits):
            splits.append(
                (
                    ts.iloc[: ((k_fold + 1) * fold_len)],
                    ts.iloc[((k_fold + 1) * fold_len) :],
                )
            )
    else:
        raise ValueError("Argument `mode` must be one of 1, 2, 3, and 4.")

    return splits



"""Module for metrics that scores predicted anomaly list against ground truth.
"""

from typing import Dict, List, Tuple, Union, overload

import pandas as pd

from ..aggregator import AndAggregator, OrAggregator
from ..data import validate_events


@overload
def recall(  # type: ignore
    y_true: pd.Series, y_pred: pd.Series, thresh: float = 0.5
) -> float:
    ...


@overload
def recall(  # type: ignore
    y_true: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    y_pred: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    thresh: float = 0.5,
) -> float:
    ...


@overload
def recall(  # type: ignore
    y_true: pd.DataFrame, y_pred: pd.DataFrame, thresh: float = 0.5
) -> Dict[str, float]:
    ...


@overload
def recall(  # type: ignore
    y_true: Dict[
        str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ],
    y_pred: Dict[
        str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ],
    thresh: float = 0.5,
) -> Dict[str, float]:
    ...


def recall(
    y_true: Union[
        pd.Series,
        pd.DataFrame,
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ],
    y_pred: Union[
        pd.Series,
        pd.DataFrame,
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ],
    thresh: float = 0.5,
) -> Union[float, Dict[str, float]]:
    """Recall score of prediction.

    Recall, a.k.a. sensitivity, hit rate, or true positive rate (TPR), is the
    percentage of true anomalies that are detected successfully.

    When the input is anomaly labels, metric calculation treats every time
    point with positive label as an independent event. An anomalous time point
    that are detected as anomalous is considered as a successful detection.

    When the input is lists of anomalous time windows, metric calculation
    treats every anomalous time window as an independent event. A true event is
    considered as successfully detected if the percentage of this time window
    covered by the detected list is greater or equal to `thresh`. Note that
    input time windows will be merged first if overlapped time windows exists
    in the list.

    Parameters
    ----------
    y_true: pandas Series or DataFrame, list, or dict
        Labels or lists of true anomalies.

        - If pandas Series, it is treated as binary labels along time index.
        - If pandas DataFrame, each column is a binary series and is treated as
          an independent type of anomaly.
        - If list, a list of events where an event is a pandas Timestamp if it
          is instantaneous or a 2-tuple of pandas Timestamps if it is a closed
          time interval.
        - If dict, each key-value pair is a list of events and is treated as an
          independent type of anomaly.

    y_pred: pandas Series or DataFrame, list, or dict
        Labels or lists of predicted anomalies.

        - If pandas Series, it is treated as binary labels along time index.
        - If pandas DataFrame, each column is a binary series and is treated as
          an independent type of anomaly.
        - If list, a list of events where an event is a pandas Timestamp if it
          is instantaneous or a 2-tuple of pandas Timestamps if it is a closed
          time interval.
        - If dict, each key-value pair is a list of events and is treated as an
          independent type of anomaly.

    thresh: float, optional
        Threshold of a hit. Only used if input is list or dict. Default: 0.5.

    Returns
    -------
    float or dict
        Score(s) for each type of anomaly.

    """
    if (thresh <= 0) or (thresh > 1):
        raise ValueError(
            "Parameter `thresh` must be a positive number less or equal to 1."
        )
    if type(y_true) != type(y_pred):
        raise TypeError("y_true and y_pred must have same type.")

    if isinstance(y_true, pd.Series) and isinstance(y_pred, pd.Series):
        try:
            pd.testing.assert_index_equal(y_true.index, y_pred.index)
        except AssertionError:
            raise ValueError("y_true and y_pred must have identical index")
        if y_true.clip(0, 1).round().sum() != 0:
            return (
                y_true.clip(0, 1).round() * y_pred.clip(0, 1).round()
            ).sum() / (y_true.clip(0, 1).round()).sum()
        else:
            return float("nan")
    elif isinstance(y_true, pd.DataFrame) and isinstance(y_pred, pd.DataFrame):
        if set(y_true.columns) != set(y_pred.columns):
            raise ValueError("y_true and y_pred must have identical columns.")
        return {
            col: recall(y_true[col], y_pred[col]) for col in y_true.columns
        }
    elif isinstance(y_true, list) and isinstance(y_pred, list):
        y_true = validate_events(y_true)
        if not y_true:
            return float("nan")
        y_int = AndAggregator().aggregate({"y_true": y_true, "y_pred": y_pred})
        n_hit = 0
        for w_true in y_true:
            if isinstance(w_true, tuple):
                true_start = w_true[0]
                true_end = w_true[1]
            else:
                true_start = w_true
                true_end = w_true
            len_w_true = true_end - true_start
            if len_w_true > pd.Timedelta(0):
                len_overlap = pd.Timedelta(0)
                for w_int in y_int:
                    if isinstance(w_int, tuple):
                        int_start = w_int[0]
                        int_end = w_int[1]
                    else:
                        int_start = w_int
                        int_end = w_int
                    len_overlap += max(
                        pd.Timedelta(0),
                        min(true_end, int_end) - max(true_start, int_start),
                    )
                if len_overlap >= thresh * len_w_true:
                    n_hit += 1
            else:
                for w_int in y_int:
                    if isinstance(w_int, tuple):
                        int_start = w_int[0]
                        int_end = w_int[1]
                    else:
                        int_start = w_int
                        int_end = w_int
                    if (int_start <= true_start) and (int_end >= true_end):
                        n_hit += 1
                        break
        return n_hit / len(y_true)
    elif isinstance(y_true, dict) and isinstance(y_pred, dict):
        return {
            key: recall(y_true[key], y_pred[key], thresh=thresh)
            for key in y_true.keys()
        }
    else:
        raise TypeError(
            "y_true and y_pred must be pandas Series or DataFrame, list, or a "
            "dict of lists"
        )


@overload
def precision(  # type: ignore
    y_true: pd.Series, y_pred: pd.Series, thresh: float = 0.5
) -> float:
    ...


@overload
def precision(  # type: ignore
    y_true: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    y_pred: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    thresh: float = 0.5,
) -> float:
    ...


@overload
def precision(  # type: ignore
    y_true: pd.DataFrame, y_pred: pd.DataFrame, thresh: float = 0.5
) -> Dict[str, float]:
    ...


@overload
def precision(  # type: ignore
    y_true: Dict[
        str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ],
    y_pred: Dict[
        str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ],
    thresh: float = 0.5,
) -> Dict[str, float]:
    ...


def precision(
    y_true: Union[
        pd.Series,
        pd.DataFrame,
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ],
    y_pred: Union[
        pd.Series,
        pd.DataFrame,
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ],
    thresh: float = 0.5,
) -> Union[float, Dict[str, float]]:
    """Precision score of prediction.

    Precision, a.k.a. positive predictive value (PPV), is the percentage of
    predicted anomalies that are true anomalies.

    When the input is anomaly labels, metric calculation treats every time
    point with positive label as an independent event. An anomalous time point
    that are detected as anomalous is considered as a successful detection.

    When the input is lists of anomalous time windows, metric calculation
    treats every anomalous time window as an independent event. A detected
    event is considered as a successfully detection if the percentage of this
    time window covered by the true anomaly list is greater or equal to
    `thresh`. Note that input time windows will be merged first if overlapped
    time windows exists in the list.

    Parameters
    ----------
    y_true: pandas Series or DataFrame, list, or dict
        Labels or lists of true anomalies.

        - If pandas Series, it is treated as binary labels along time index.
        - If pandas DataFrame, each column is a binary series and is treated as
          an independent type of anomaly.
        - If list, a list of events where an event is a pandas Timestamp if it
          is instantaneous or a 2-tuple of pandas Timestamps if it is a closed
          time interval.
        - If dict, each key-value pair is a list of events and is treated as an
          independent type of anomaly.

    y_pred: pandas Series or DataFrame, list, or dict
        Labels or lists of predicted anomalies.

        - If pandas Series, it is treated as binary labels along time index.
        - If pandas DataFrame, each column is a binary series and is treated as
          an independent type of anomaly.
        - If list, a list of events where an event is a pandas Timestamp if it
          is instantaneous or a 2-tuple of pandas Timestamps if it is a closed
          time interval.
        - If dict, each key-value pair is a list of events and is treated as an
          independent type of anomaly.

    thresh: float, optional
        Threshold of a hit. Only used if input is list or dict. Default: 0.5.

    Returns
    -------
    float or dict
        Score(s) for each type of anomaly.

    """
    return recall(y_pred, y_true, thresh=thresh)


@overload
def f1_score(  # type: ignore
    y_true: pd.Series,
    y_pred: pd.Series,
    recall_thresh: float = 0.5,
    precision_thresh: float = 0.5,
) -> float:
    ...


@overload
def f1_score(  # type: ignore
    y_true: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    y_pred: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    recall_thresh: float = 0.5,
    precision_thresh: float = 0.5,
) -> float:
    ...


@overload
def f1_score(  # type: ignore
    y_true: pd.DataFrame,
    y_pred: pd.DataFrame,
    recall_thresh: float = 0.5,
    precision_thresh: float = 0.5,
) -> Dict[str, float]:
    ...


@overload
def f1_score(  # type: ignore
    y_true: Dict[
        str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ],
    y_pred: Dict[
        str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ],
    recall_thresh: float = 0.5,
    precision_thresh: float = 0.5,
) -> Dict[str, float]:
    ...


def f1_score(
    y_true: Union[
        pd.Series,
        pd.DataFrame,
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ],
    y_pred: Union[
        pd.Series,
        pd.DataFrame,
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ],
    recall_thresh: float = 0.5,
    precision_thresh: float = 0.5,
) -> Union[float, Dict[str, float]]:
    """F1 score of prediction.

    F1 score is the harmonic mean of precision and recall. For more details
    about precision and recall, please refer to function `precision` and
    `recall`.

    Parameters
    ----------
    y_true: pandas Series or DataFrame, list, or dict
        Labels or lists of true anomalies.

        - If pandas Series, it is treated as binary labels along time index.
        - If pandas DataFrame, each column is a binary series and is treated as
          an independent type of anomaly.
        - If list, a list of events where an event is a pandas Timestamp if it
          is instantaneous or a 2-tuple of pandas Timestamps if it is a closed
          time interval.
        - If dict, each key-value pair is a list of events and is treated as an
          independent type of anomaly.

    y_pred: pandas Series or DataFrame, list, or dict
        Labels or lists of predicted anomalies.

        - If pandas Series, it is treated as binary labels along time index.
        - If pandas DataFrame, each column is a binary series and is treated as
          an independent type of anomaly.
        - If list, a list of events where an event is a pandas Timestamp if it
          is instantaneous or a 2-tuple of pandas Timestamps if it is a closed
          time interval.
        - If dict, each key-value pair is a list of events and is treated as an
          independent type of anomaly.

    recall_thresh: float, optional
        Threshold of recall calculation. Only used if input is list or dict.
        For more details, please refer to function `recall`. Default: 0.5.

    precision_thresh: float, optional
        Threshold of precision calculation. Only used if input is list or dict.
        For more details, please refer to function `precision`. Default: 0.5.

    Returns
    -------
    float or dict
        Score(s) for each type of anomaly.

    """
    recall_score = recall(y_true, y_pred, recall_thresh)
    precision_score = precision(y_true, y_pred, precision_thresh)
    if isinstance(recall_score, float) and isinstance(precision_score, float):
        if recall_score + precision_score != 0:
            return (
                2
                * recall_score
                * precision_score
                / (recall_score + precision_score)
            )
        else:
            return float("nan")
    elif isinstance(recall_score, dict) and isinstance(precision_score, dict):
        return {
            key: (
                (
                    2
                    * recall_score[key]
                    * precision_score[key]
                    / (recall_score[key] + precision_score[key])
                )
                if (recall_score[key] + precision_score[key] != 0)
                else float("nan")
            )
            for key in recall_score.keys()
        }
    else:
        raise RuntimeError("This error is not supposed to be hit.")


@overload
def iou(  # type: ignore
    y_true: pd.Series, y_pred: pd.Series
) -> float:
    ...


@overload
def iou(  # type: ignore
    y_true: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
    y_pred: List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
) -> float:
    ...


@overload
def iou(  # type: ignore
    y_true: pd.DataFrame, y_pred: pd.DataFrame
) -> Dict[str, float]:
    ...


@overload
def iou(  # type: ignore
    y_true: Dict[
        str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ],
    y_pred: Dict[
        str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ],
) -> Dict[str, float]:
    ...


def iou(
    y_true: Union[
        pd.Series,
        pd.DataFrame,
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ],
    y_pred: Union[
        pd.Series,
        pd.DataFrame,
        List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
        Dict[
            str, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
        ],
    ],
) -> Union[float, Dict[str, float]]:
    """IoU (Intersection over Union) score of prediction.

    Intersection over union is the length ratio between time segments that are
    identified as anomalous in both lists and those identified by at least one
    of the two lists.

    When the input is anomaly labels, metric calculation counts the number of
    anomalous time points.

    When the input is lists of anomalous time windows, metric calculation
    measure the length of time segments.

    Parameters
    ----------
    y_true: pandas Series or DataFrame, list, or dict
        Labels or lists of true anomalies.

        - If pandas Series, it is treated as binary labels along time index.
        - If pandas DataFrame, each column is a binary series and is treated as
          an independent type of anomaly.
        - If list, a list of events where an event is a pandas Timestamp if it
          is instantaneous or a 2-tuple of pandas Timestamps if it is a closed
          time interval.
        - If dict, each key-value pair is a list of events and is treated as an
          independent type of anomaly.

    y_pred: pandas Series or DataFrame, list, or dict
        Labels or lists of predicted anomalies.

        - If pandas Series, it is treated as binary labels along time index.
        - If pandas DataFrame, each column is a binary series and is treated as
          an independent type of anomaly.
        - If list, a list of events where an event is a pandas Timestamp if it
          is instantaneous or a 2-tuple of pandas Timestamps if it is a closed
          time interval.
        - If dict, each key-value pair is a list of events and is treated as an
          independent type of anomaly.

    Returns
    -------
    float or dict
        Score(s) for each type of anomaly.

    """
    if type(y_true) != type(y_pred):
        raise TypeError("y_true and y_pred must have same type.")

    if isinstance(y_true, pd.Series) and isinstance(y_pred, pd.Series):
        try:
            pd.testing.assert_index_equal(y_true.index, y_pred.index)
        except AssertionError:
            raise ValueError("y_true and y_pred must have identical index")
        if (y_true.clip(0, 1).round() + y_pred.clip(0, 1).round()).clip(
            0, 1
        ).sum() != 0:
            return (
                (y_true.clip(0, 1).round() * y_pred.clip(0, 1).round()).sum()
                / (y_true.clip(0, 1).round() + y_pred.clip(0, 1).round())
                .clip(0, 1)
                .sum()
            )
        else:
            return float("nan")
    elif isinstance(y_true, pd.DataFrame) and isinstance(y_pred, pd.DataFrame):
        if set(y_true.columns) != set(y_pred.columns):
            raise ValueError("y_true and y_pred must have identical columns.")
        return {col: iou(y_true[col], y_pred[col]) for col in y_true.columns}
    elif isinstance(y_true, list) and isinstance(y_pred, list):
        y_int = AndAggregator().aggregate({"y_true": y_true, "y_pred": y_pred})
        y_union = OrAggregator().aggregate(
            {"y_true": y_true, "y_pred": y_pred}
        )
        len_int = sum(
            [
                (w[1] - w[0]).total_seconds() if isinstance(w, tuple) else 0
                for w in y_int
            ]
        )
        len_union = sum(
            [
                (w[1] - w[0]).total_seconds() if isinstance(w, tuple) else 0
                for w in y_union
            ]
        )
        if len_union == 0:
            return float("nan")
        return len_int / len_union
    elif isinstance(y_true, dict) and isinstance(y_pred, dict):
        return {key: iou(y_true[key], y_pred[key]) for key in y_true.keys()}
    else:
        raise TypeError(
            "y_true and y_pred must be pandas Series or DataFrame, list, or a "
            "dict of lists"
        )



"""Module of model pipeline and pipenet.

Pipeline or Pipenet connects multiple components (transformers, detectors,
and/or aggregators) into a model that may perform complex anomaly detection
process.

"""

from ._pipe import Pipeline, Pipenet

__all__ = ["Pipeline", "Pipenet"]



"""Module for 1-dimensional detectors.

1-dimensional detectors detect anomalies from 1-dimensional time series, i.e.
from pandas Series.

"""

from typing import Any, Callable, Dict, Optional, Tuple, Union

import numpy as np
import pandas as pd
from scipy.stats import t
from sklearn.linear_model import LinearRegression

from .._detector_base import (
    _NonTrainableUnivariateDetector,
    _TrainableUnivariateDetector,
)
from ..aggregator import AndAggregator
from ..pipe import Pipenet
from ..transformer import (
    ClassicSeasonalDecomposition,
    CustomizedTransformer1D,
    DoubleRollingAggregate,
    RegressionResidual,
    Retrospect,
)


class CustomizedDetector1D(_TrainableUnivariateDetector):
    """Univariate detector derived from a user-given function and parameters.

    Parameters
    ----------
    detect_func: function
        A function detecting anomalies from univariate time series.

        The first input argument must be a pandas Series, optional input
        argument may be accepted through parameter `detect_func_params` and the
        output of `fit_func`, and the output must be a binary pandas Series
        with the same index as input.

    detect_func_params: dict, optional
        Parameters of `detect_func`. Default: None.

    fit_func: function, optional
        A function training parameters of `detect_func` with univariate time
        series.

        The first input argument must be a pandas Series, optional input
        argument may be accepted through parameter `fit_func_params`, and the
        output must be a dict that can be used by `detect_func` as parameters.
        Default: None.

    fit_func_params: dict, optional
        Parameters of `fit_func`. Default: None.

    """

    def __init__(
        self,
        detect_func: Callable,
        detect_func_params: Optional[Dict[str, Any]] = None,
        fit_func: Optional[Callable] = None,
        fit_func_params: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._fitted_detect_func_params = {}  # type: Dict[str, Any]
        super().__init__()
        self.detect_func = detect_func
        self.detect_func_params = detect_func_params
        self.fit_func = fit_func
        self.fit_func_params = fit_func_params
        if self.fit_func is None:
            self._fitted = 1

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return (
            "detect_func",
            "detect_func_params",
            "fit_func",
            "fit_func_params",
        )

    def _fit_core(self, s: pd.Series) -> None:
        if self.fit_func is not None:
            if self.fit_func_params is not None:
                fit_func_params = self.fit_func_params
            else:
                fit_func_params = {}
            self._fitted_detect_func_params = self.fit_func(
                s, **fit_func_params
            )

    def _predict_core(self, s: pd.Series) -> pd.Series:
        if self.detect_func_params is not None:
            detect_func_params = self.detect_func_params
        else:
            detect_func_params = {}
        if self.fit_func is not None:
            return self.detect_func(
                s, **{**self._fitted_detect_func_params, **detect_func_params}
            )
        else:
            return self.detect_func(s, **detect_func_params)


class ThresholdAD(_NonTrainableUnivariateDetector):
    """Detector that detects anomaly based on user-given threshold.

    This detector compares time series values with user-given thresholds, and
    identifies time points as anomalous when values are beyond the thresholds.

    Parameters
    ----------
    low: float, optional
        Threshold below which a value is regarded anomaly. Default: None, i.e.
        no threshold on lower side.

    high: float, optional
        Threshold above which a value is regarded anomaly. Default: None, i.e.
        no threshold on upper side.

    """

    def __init__(
        self, low: Optional[float] = None, high: Optional[float] = None
    ) -> None:
        super().__init__()
        self.low = low
        self.high = high

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("low", "high")

    def _predict_core(self, s: pd.Series) -> pd.Series:
        predicted = (
            s > (self.high if (self.high is not None) else float("inf"))
        ) | (s < (self.low if (self.low is not None) else -float("inf")))
        predicted[s.isna()] = np.nan
        return predicted


class QuantileAD(_TrainableUnivariateDetector):
    """Detector that detects anomaly based on quantiles of historical data.

    This detector compares time series values with user-specified quantiles
    of historical data, and identifies time points as anomalous when values
    are beyond the thresholds.

    Parameters
    ----------
    low: float, optional
        Quantile of historical data lower which a value is regarded as anomaly.
        Must between 0 and 1. Default: None, i.e. no threshold on lower side.

    high: float, optional
        Quantile of historical data above which a value is regarded as anomaly.
        Must between 0 and 1. Default: None, i.e. no threshold on upper side.

    Attributes
    ----------
    abs_low_: float
        The fitted lower bound of normal range.

    abs_high_: float
        The fitted upper bound of normal range.

    """

    def __init__(
        self, low: Optional[float] = None, high: Optional[float] = None
    ) -> None:
        super().__init__()
        self.low = low
        self.high = high

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("low", "high")

    def _fit_core(self, s: pd.Series) -> None:
        if s.count() == 0:
            raise RuntimeError("Valid values are not enough for training.")
        if self.high is None:
            self.abs_high_ = float("inf")
        else:
            self.abs_high_ = s.quantile(self.high)
        if self.low is None:
            self.abs_low_ = -float("inf")
        else:
            self.abs_low_ = s.quantile(self.low)

    def _predict_core(self, s: pd.Series) -> pd.Series:
        predicted = (s > self.abs_high_) | (s < self.abs_low_)
        predicted[s.isna()] = np.nan
        return predicted


class InterQuartileRangeAD(_TrainableUnivariateDetector):
    """
    Detector that detects anomaly based on inter-quartile range of historical
    data.

    This detector compares time series values with 1st and 3rd quartiles of
    historical data, and identifies time points as anomalous when differences
    are beyond the inter-quartile range (IQR) times a user-given factor c.

    Parameters
    ----------
    c: float, or 2-tuple (float, float), optional
        Factor used to determine the bound of normal range (betweeen Q1-c*IQR
        and Q3+c*IQR). If a tuple (c1, c2), the factors are for lower and upper
        bound respectively. Default: 3.0.

    Attributes
    ----------
    abs_low_: float
        The fitted lower bound of normal range.

    abs_high_: float
        The fitted upper bound of normal range.

    """

    def __init__(
        self,
        c: Union[
            Optional[float], Tuple[Optional[float], Optional[float]]
        ] = 3.0,
    ) -> None:
        super().__init__()
        self.c = c

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("c",)

    def _fit_core(self, s: pd.Series) -> None:
        if s.count() == 0:
            raise RuntimeError("Valid values are not enough for training.")
        q1 = s.quantile(0.25)
        q3 = s.quantile(0.75)
        iqr = q3 - q1
        self.abs_low_ = (
            (
                q1
                - iqr
                * (self.c if (not isinstance(self.c, tuple)) else self.c[0])
            )
            if (
                (self.c if (not isinstance(self.c, tuple)) else self.c[0])
                is not None
            )
            else -float("inf")
        )
        self.abs_high_ = (
            q3
            + iqr * (self.c if (not isinstance(self.c, tuple)) else self.c[1])
            if (
                (self.c if (not isinstance(self.c, tuple)) else self.c[1])
                is not None
            )
            else float("inf")
        )

    def _predict_core(self, s: pd.Series) -> pd.Series:
        predicted = (s > self.abs_high_) | (s < self.abs_low_)
        predicted[s.isna()] = np.nan
        return predicted


class GeneralizedESDTestAD(_TrainableUnivariateDetector):
    """Detector that detects anomaly based on generalized ESD test.

    This detector performs generalized extreme Studentized deviate (ESD) test
    [1, 2] on historical data and identifies normal values vs. outliers for
    training. For predicting, the detector adds each value in the testing
    series to the set of normal values from training series independently, and
    performs generalized ESD test to this set (all normal values from training
    series, plus one value from testing series) to evaluate if this value of
    interest is an outlier.

    Please note a key assumption of generalized ESD test is that values follow
    an approximately normal distribution. Please only use this detector when
    this assumption holds.

    [1] Rosner, Bernard (May 1983), Percentage Points for a Generalized ESD
    Many-Outlier Procedure,Technometrics, 25(2), pp. 165-172.

    [2] https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm

    Parameters
    ----------
    alpha: float, optional
        Significance level. Default: 0.05.

    """

    def __init__(self, alpha: float = 0.05) -> None:
        super().__init__()
        self.alpha = alpha

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("alpha",)

    def _fit_core(self, s: pd.Series) -> None:
        if s.count() == 0:
            raise RuntimeError("Valid values are not enough for training.")
        R = pd.Series(np.zeros(len(s)), index=s.index)
        n = s.count()
        Lambda = pd.Series(np.zeros(len(s)), index=s.index)
        s_copy = s.copy()
        i = 0
        while s_copy.count() > 0:
            i += 1
            ind = (s_copy - s_copy.mean()).abs().idxmax()
            R[ind] = (
                abs(s_copy[ind] - s_copy.mean()) / s_copy.std()
                if s_copy.std() > 0
                else 0
            )
            s_copy[ind] = np.nan
            p = 1 - self.alpha / (2 * (n - i + 1))
            Lambda[ind] = (
                (n - i)
                * t.ppf(p, n - i - 1)
                / np.sqrt((n - i - 1 + t.ppf(p, n - i - 1) ** 2) * (n - i + 1))
            )
            if R[ind] <= Lambda[ind]:
                break
        self._normal_sum = s[Lambda >= R].sum()
        self._normal_squared_sum = (s[Lambda >= R] ** 2).sum()
        self._normal_count = s[Lambda >= R].count()
        i = 1
        n = self._normal_count + 1
        p = 1 - self.alpha / (2 * (n - i + 1))
        self._lambda = (
            (n - i)
            * t.ppf(p, n - i - 1)
            / np.sqrt((n - i - 1 + t.ppf(p, n - i - 1) ** 2) * (n - i + 1))
        )

    def _predict_core(self, s: pd.Series) -> pd.Series:
        new_sum = s + self._normal_sum
        new_count = self._normal_count + 1
        new_mean = new_sum / new_count
        new_squared_sum = s ** 2 + self._normal_squared_sum
        new_std = np.sqrt(
            (
                new_squared_sum
                - 2 * new_mean * new_sum
                + new_count * new_mean ** 2
            )
            / (new_count - 1)
        )
        predicted = (s - new_mean).abs() / new_std > self._lambda
        predicted[s.isna()] = np.nan
        return predicted


# =============================================================================
# PLEASE PUT PIPE-DERIVED DETECTOR CLASSES BELOW THIS LINE
# =============================================================================


class PersistAD(_TrainableUnivariateDetector):
    """Detector that detects anomaly based on values in a preceding period.

    This detector compares time series values with the values of their
    preceding time windows, and identifies a time point as anomalous if the
    change of value from its preceding average or median is anomalously large.

    This detector is internally implemented as a `Pipenet` object. Advanced
    users may learn more details by checking attribute `pipe_`.

    Parameters
    ----------
    window: int or str, optional
        Size of the preceding time window.

        - If int, it is the number of time point in this time window.
        - If str, it must be able to be converted into a pandas Timedelta
          object.

        Default: 1.

    c: float, optional
        Factor used to determine the bound of normal range based on historical
        interquartile range. Default: 3.0.

    side: str, optional
        - If "both", to detect anomalous positive and negative changes;
        - If "positive", to only detect anomalous positive changes;
        - If "negative", to only detect anomalous negative changes.

        Default: "both".

    min_periods: int, optional
        Minimum number of observations in each window required to have a value
        for that window. Default: None, i.e. all observations must have values.

    agg: str, optional
        Aggregation operation of the time window, either "mean" or "median".
        Default: "median".

    Attributes
    ----------
    pipe_: adtk.pipe.Pipenet
        Internal pipenet object.

    """

    def __init__(
        self,
        window: Union[int, str] = 1,
        c: float = 3.0,
        side: str = "both",
        min_periods: Optional[int] = None,
        agg: str = "median",
    ) -> None:
        self.pipe_ = Pipenet(
            {
                "diff_abs": {
                    "model": DoubleRollingAggregate(
                        agg=agg,
                        window=(window, 1),
                        center=True,
                        min_periods=(min_periods, 1),
                        diff="l1",
                    ),
                    "input": "original",
                },
                "iqr_ad": {
                    "model": InterQuartileRangeAD(c=(None, c)),
                    "input": "diff_abs",
                },
                "diff": {
                    "model": DoubleRollingAggregate(
                        agg=agg,
                        window=(window, 1),
                        center=True,
                        min_periods=(min_periods, 1),
                        diff="diff",
                    ),
                    "input": "original",
                },
                "sign_check": {
                    "model": ThresholdAD(
                        high=(
                            0.0
                            if side == "positive"
                            else (
                                float("inf")
                                if side == "negative"
                                else -float("inf")
                            )
                        ),
                        low=(
                            0.0
                            if side == "negative"
                            else (
                                -float("inf")
                                if side == "positive"
                                else float("inf")
                            )
                        ),
                    ),
                    "input": "diff",
                },
                "and": {
                    "model": AndAggregator(),
                    "input": ["iqr_ad", "sign_check"],
                },
            }
        )
        super().__init__()
        self.c = c
        self.side = side
        self.window = window
        self.min_periods = min_periods
        self.agg = agg
        self._sync_params()

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("window", "c", "side", "min_periods", "agg")

    def _sync_params(self) -> None:
        if self.agg not in ["median", "mean"]:
            raise ValueError(
                "Parameter `agg` must be either 'median' or 'mean'."
            )
        if self.side not in ["both", "positive", "negative"]:
            raise ValueError(
                "Parameter `side` must be 'both', 'positive' or 'negative'."
            )
        self.pipe_.steps["diff_abs"]["model"].set_params(
            agg=self.agg,
            window=(self.window, 1),
            min_periods=(self.min_periods, 1),
        )
        self.pipe_.steps["iqr_ad"]["model"].set_params(c=(None, self.c))
        self.pipe_.steps["diff"]["model"].set_params(
            agg=self.agg,
            window=(self.window, 1),
            min_periods=(self.min_periods, 1),
        )
        self.pipe_.steps["sign_check"]["model"].set_params(
            high=(
                0.0
                if self.side == "positive"
                else (
                    float("inf") if self.side == "negative" else -float("inf")
                )
            ),
            low=(
                0.0
                if self.side == "negative"
                else (
                    -float("inf") if self.side == "positive" else float("inf")
                )
            ),
        )

    def _fit_core(self, s: pd.Series) -> None:
        self._sync_params()
        self.pipe_.fit(s)

    def _predict_core(self, s: pd.Series) -> pd.Series:
        self._sync_params()
        return self.pipe_.detect(s)


class LevelShiftAD(_TrainableUnivariateDetector):
    """Detector that detects level shift of time series values.

    This detector compares values of two time windows next to each others, and
    identifies the time point in between as an level-shift point if the
    difference of the medians in the two time windows is anomalously large.

    This detector is internally implemented as a `Pipenet` object. Advanced
    users may learn more details by checking attribute `pipe_`.

    Parameters
    ----------
    window: int or str, or 2-tuple of int or str
        Size of the time windows.

        - If int, it is the number of time point in this time window.
        - If str, it must be able to be converted into a pandas Timedelta
          object.
        - If 2-tuple, it defines the left and right window respectively.

    c: float, optional
        Factor used to determine the bound of normal range based on historical
        interquartile range. Default: 6.0.

    side: str, optional
        - If "both", to detect anomalous positive and negative changes;
        - If "positive", to only detect anomalous positive changes;
        - If "negative", to only detect anomalous negative changes.

        Default: "both".

    min_periods: int, or 2-tuple of int, optional
        Minimum number of observations in each window required to have a value
        for that window. If 2-tuple, it defines the left and right window
        respectively. Default: None, i.e. all observations must have values.

    Attributes
    ----------
    pipe_: adtk.pipe.Pipenet
        Internal pipenet object.

    """

    def __init__(
        self,
        window: Union[
            Union[int, str], Tuple[Union[int, str], Union[int, str]]
        ],
        c: float = 6.0,
        side: str = "both",
        min_periods: Union[
            Optional[int], Tuple[Optional[int], Optional[int]]
        ] = None,
    ) -> None:
        self.pipe_ = Pipenet(
            {
                "diff_abs": {
                    "model": DoubleRollingAggregate(
                        agg="median",
                        window=window,
                        center=True,
                        min_periods=min_periods,
                        diff="l1",
                    ),
                    "input": "original",
                },
                "iqr_ad": {
                    "model": InterQuartileRangeAD((None, c)),
                    "input": "diff_abs",
                },
                "diff": {
                    "model": DoubleRollingAggregate(
                        agg="median",
                        window=window,
                        center=True,
                        min_periods=min_periods,
                        diff="diff",
                    ),
                    "input": "original",
                },
                "sign_check": {
                    "model": ThresholdAD(
                        high=(
                            0.0
                            if side == "positive"
                            else (
                                float("inf")
                                if side == "negative"
                                else -float("inf")
                            )
                        ),
                        low=(
                            0.0
                            if side == "negative"
                            else (
                                -float("inf")
                                if side == "positive"
                                else float("inf")
                            )
                        ),
                    ),
                    "input": "diff",
                },
                "and": {
                    "model": AndAggregator(),
                    "input": ["iqr_ad", "sign_check"],
                },
            }
        )
        super().__init__()
        self.c = c
        self.side = side
        self.window = window
        self.min_periods = min_periods
        self._sync_params()

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("window", "c", "side", "min_periods")

    def _sync_params(self) -> None:
        if self.side not in ["both", "positive", "negative"]:
            raise ValueError(
                "Parameter `side` must be 'both', 'positive' or 'negative'."
            )
        self.pipe_.steps["diff_abs"]["model"].set_params(
            window=self.window, min_periods=self.min_periods
        )
        self.pipe_.steps["iqr_ad"]["model"].set_params(c=(None, self.c))
        self.pipe_.steps["diff"]["model"].set_params(
            window=self.window, min_periods=self.min_periods
        )
        self.pipe_.steps["sign_check"]["model"].set_params(
            high=(
                0.0
                if self.side == "positive"
                else (
                    float("inf") if self.side == "negative" else -float("inf")
                )
            ),
            low=(
                0.0
                if self.side == "negative"
                else (
                    -float("inf") if self.side == "positive" else float("inf")
                )
            ),
        )

    def _fit_core(self, s: pd.Series) -> None:
        self._sync_params()
        self.pipe_.fit(s)

    def _predict_core(self, s: pd.Series) -> pd.Series:
        self._sync_params()
        return self.pipe_.detect(s)


class VolatilityShiftAD(_TrainableUnivariateDetector):
    """Detector that detects shift of volatility in time series.

    This detector compares volatility of two time windows next to each others,
    and identifies the time point in between as a volatility-shift point if the
    difference of the volatility measurement in the two time windows is
    anomalously large.

    This detector is internally implemented as a `Pipenet` object. Advanced
    users may learn more details by checking attribute `pipe_`.

    Parameters
    ----------
    window: int or str, or 2-tuple of int or str
        Size of the time windows.

        - If int, it is the number of time point in this time window.
        - If str, it must be able to be converted into a pandas Timedelta
          object.
        - If 2-tuple, it defines the left and right window respectively.

    c: float, optional
        Factor used to determine the bound of normal range based on historical
        interquartile range. Default: 6.0.

    side: str, optional
        - If "both", to detect anomalous positive and negative changes;
        - If "positive", to only detect anomalous positive changes;
        - If "negative", to only detect anomalous negative changes.

        Default: "both".

    min_periods: int, optional
        Minimum number of observations in each window required to have a value
        for that window. If 2-tuple, it defines the left and right window
        respectively. Default: None, i.e. all observations must have values.

    agg: str, optional
        Measurement of volatility in a time window, one of "std" (standard
        deviation), "iqr" (interquartile range), or "idr" (interdecile range).
        Default: "std".

    Attributes
    ----------
    pipe_: adtk.pipe.Pipenet
        Internal pipenet object.

    """

    def __init__(
        self,
        window: Union[
            Union[int, str], Tuple[Union[int, str], Union[int, str]]
        ],
        c: float = 6.0,
        side: str = "both",
        min_periods: Union[
            Optional[int], Tuple[Optional[int], Optional[int]]
        ] = None,
        agg: str = "std",
    ) -> None:
        self.pipe_ = Pipenet(
            {
                "diff_abs": {
                    "model": DoubleRollingAggregate(
                        agg=agg,
                        window=window,
                        center=True,
                        min_periods=min_periods,
                        diff="abs_rel_diff",
                    ),
                    "input": "original",
                },
                "iqr_ad": {
                    "model": InterQuartileRangeAD((None, c)),
                    "input": "diff_abs",
                },
                "diff": {
                    "model": DoubleRollingAggregate(
                        agg=agg,
                        window=window,
                        center=True,
                        min_periods=min_periods,
                        diff="diff",
                    ),
                    "input": "original",
                },
                "sign_check": {
                    "model": ThresholdAD(
                        high=(
                            0.0
                            if side == "positive"
                            else (
                                float("inf")
                                if side == "negative"
                                else -float("inf")
                            )
                        ),
                        low=(
                            0.0
                            if side == "negative"
                            else (
                                -float("inf")
                                if side == "positive"
                                else float("inf")
                            )
                        ),
                    ),
                    "input": "diff",
                },
                "and": {
                    "model": AndAggregator(),
                    "input": ["iqr_ad", "sign_check"],
                },
            }
        )
        super().__init__()
        self.agg = agg
        self.c = c
        self.side = side
        self.window = window
        self.min_periods = min_periods
        self._sync_params()

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("window", "c", "side", "min_periods", "agg")

    def _sync_params(self) -> None:
        if self.agg not in ["std", "iqr", "idr"]:
            raise ValueError("Parameter `agg` must be 'std', 'iqr' or 'idr'.")
        if self.side not in ["both", "positive", "negative"]:
            raise ValueError(
                "Parameter `side` must be 'both', 'positive' or 'negative'."
            )
        self.pipe_.steps["diff_abs"]["model"].set_params(
            agg=self.agg, window=self.window, min_periods=self.min_periods
        )
        self.pipe_.steps["iqr_ad"]["model"].set_params(c=(None, self.c))
        self.pipe_.steps["diff"]["model"].set_params(
            agg=self.agg, window=self.window, min_periods=self.min_periods
        )
        self.pipe_.steps["sign_check"]["model"].set_params(
            high=(
                0.0
                if self.side == "positive"
                else (
                    float("inf") if self.side == "negative" else -float("inf")
                )
            ),
            low=(
                0.0
                if self.side == "negative"
                else (
                    -float("inf") if self.side == "positive" else float("inf")
                )
            ),
        )

    def _fit_core(self, s: pd.Series) -> None:
        self._sync_params()
        self.pipe_.fit(s)

    def _predict_core(self, s: pd.Series) -> pd.Series:
        self._sync_params()
        return self.pipe_.detect(s)


class AutoregressionAD(_TrainableUnivariateDetector):
    """Detector that detects anomalous autoregression property in time series.

    Many time series has autoregressive behavior. For example, in a linearly
    autoregressive time series, current value is a linear combination of
    serveral previous values. Violation of usual autoregressive behavior may
    indicate anomaly.

    The detector applies a regressor to learn autoregressive property of the
    time series, and identifies a time point as anomalous when the residual of
    autoregression is anomalously large.

    This detector is internally implemented as a `Pipenet` object. Advanced
    users may learn more details by checking attribute `pipe_`.

    Parameters
    ----------
    n_steps: int, optional
        Number of steps (previous values) to include in the model. Default: 1.

    step_size: int, optional
        Length of a step. For example, if n_steps=2, step_size=3, X_[t-3] and
        X_[t-6] will be used to predict X_[t]. Default: 1.

    regressor: object, optional
        Regressor to be used. Same as a scikit-learn regressor, it should
        minimally have `fit` and `predict` methods. If not given, a linear
        regressor will be used.

    c: float, optional
        Factor used to determine the bound of normal range based on historical
        interquartile range. Default: 3.0.

    side: str, optional
        - If "both", to detect anomalous positive and negative residuals;
        - If "positive", to only detect anomalous positive residuals;
        - If "negative", to only detect anomalous negative residuals.

        Default: "both".

    Attributes
    ----------
    pipe_: adtk.pipe.Pipenet
        Internal pipenet object.

    """

    def __init__(
        self,
        n_steps: int = 1,
        step_size: int = 1,
        regressor: Optional[Any] = None,
        c: float = 3.0,
        side: str = "both",
    ) -> None:
        if regressor is None:
            regressor = LinearRegression()
        self.pipe_ = Pipenet(
            {
                "retrospetive": {
                    "model": Retrospect(
                        n_steps=n_steps + 1, step_size=step_size
                    ),
                    "input": "original",
                },
                "regression_residual": {
                    "model": RegressionResidual(
                        target="t-0", regressor=regressor
                    ),
                    "input": "retrospetive",
                },
                "abs_residual": {
                    "model": CustomizedTransformer1D(transform_func=abs),
                    "input": "regression_residual",
                },
                "iqr_ad": {
                    "model": InterQuartileRangeAD((None, c)),
                    "input": "abs_residual",
                },
                "sign_check": {
                    "model": ThresholdAD(
                        high=(
                            0.0
                            if side == "positive"
                            else (
                                float("inf")
                                if side == "negative"
                                else -float("inf")
                            )
                        ),
                        low=(
                            0.0
                            if side == "negative"
                            else (
                                -float("inf")
                                if side == "positive"
                                else float("inf")
                            )
                        ),
                    ),
                    "input": "regression_residual",
                },
                "and": {
                    "model": AndAggregator(),
                    "input": ["iqr_ad", "sign_check"],
                },
            }
        )
        super().__init__()
        self.n_steps = n_steps
        self.step_size = step_size
        self.regressor = regressor
        self.c = c
        self.side = side
        self._sync_params()

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("n_steps", "step_size", "regressor", "c", "side")

    def _sync_params(self) -> None:
        if self.side not in ["both", "positive", "negative"]:
            raise ValueError(
                "Parameter `side` must be 'both', 'positive' or 'negative'."
            )
        if self.regressor is None:
            self.regressor = LinearRegression()
        self.pipe_.steps["retrospetive"]["model"].set_params(
            n_steps=self.n_steps + 1, step_size=self.step_size
        )
        self.pipe_.steps["regression_residual"]["model"].set_params(
            regressor=self.regressor
        )
        self.pipe_.steps["iqr_ad"]["model"].set_params(c=(None, self.c))
        self.pipe_.steps["sign_check"]["model"].set_params(
            high=(
                0.0
                if self.side == "positive"
                else (
                    float("inf") if self.side == "negative" else -float("inf")
                )
            ),
            low=(
                0.0
                if self.side == "negative"
                else (
                    -float("inf") if self.side == "positive" else float("inf")
                )
            ),
        )

    def _fit_core(self, s: pd.Series) -> None:
        self._sync_params()
        self.pipe_.fit(s)

    def _predict_core(self, s: pd.Series) -> pd.Series:
        self._sync_params()
        return self.pipe_.detect(s)


class SeasonalAD(_TrainableUnivariateDetector):
    """Detector that detects anomalous values away from seasonal pattern.

    This detector uses a seasonal decomposition transformer to remove seasonal
    pattern (as well as trend optional), and identifies a time point as
    anomalous when the residual of seasonal decomposition is anomalously large.

    This detector is internally implemented as a `Pipenet` object. Advanced
    users may learn more details by checking attribute `pipe_`.

    Parameters
    ----------
    freq: int, optional
        Length of a seasonal cycle as the number of time points in a cycle. If
        not specified, the model will try to determine it based on
        autocorrelation of the training series. Default: None.

    c: float, optional
        Factor used to determine the bound of normal range based on historical
        interquartile range. Default: 3.0.

    side: str, optional
        - If "both", to detect anomalous positive and negative residuals;
        - If "positive", to only detect anomalous positive residuals;
        - If "negative", to only detect anomalous negative residuals.

        Default: "both".

    trend: bool, optional
        Whether to extract trend during decomposition. Default: False.

    Attributes
    ----------
    freq_: int
        Length of seasonal cycle as the number of time points in a cycle. Equal
        to parameter `freq` if it is specified. Otherwise, calculated based on
        autocorrelation of the training series.

    seasonal_: pandas.Series
        Seasonal pattern extracted from training series.

    pipe_: adtk.pipe.Pipenet
        Internal pipenet object.

    """

    def __init__(
        self,
        freq: Optional[int] = None,
        side: str = "both",
        c: float = 3.0,
        trend: bool = False,
    ) -> None:
        self.pipe_ = Pipenet(
            {
                "deseasonal_residual": {
                    "model": (
                        ClassicSeasonalDecomposition(freq=freq, trend=trend)
                    ),
                    "input": "original",
                },
                "abs_residual": {
                    "model": CustomizedTransformer1D(transform_func=abs),
                    "input": "deseasonal_residual",
                },
                "iqr_ad": {
                    "model": InterQuartileRangeAD((None, c)),
                    "input": "abs_residual",
                },
                "sign_check": {
                    "model": ThresholdAD(
                        high=(
                            0.0
                            if side == "positive"
                            else (
                                float("inf")
                                if side == "negative"
                                else -float("inf")
                            )
                        ),
                        low=(
                            0.0
                            if side == "negative"
                            else (
                                -float("inf")
                                if side == "positive"
                                else float("inf")
                            )
                        ),
                    ),
                    "input": "deseasonal_residual",
                },
                "and": {
                    "model": AndAggregator(),
                    "input": ["iqr_ad", "sign_check"],
                },
            }
        )
        super().__init__()
        self.freq = freq
        self.side = side
        self.c = c
        self.trend = trend
        self._sync_params()

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("freq", "side", "c", "trend")

    def _sync_params(self) -> None:
        self.pipe_.steps["deseasonal_residual"]["model"].set_params(
            freq=self.freq, trend=self.trend
        )
        self.pipe_.steps["iqr_ad"]["model"].set_params(c=(None, self.c))
        self.pipe_.steps["sign_check"]["model"].set_params(
            high=(
                0.0
                if self.side == "positive"
                else (
                    float("inf") if self.side == "negative" else -float("inf")
                )
            ),
            low=(
                0.0
                if self.side == "negative"
                else (
                    -float("inf") if self.side == "positive" else float("inf")
                )
            ),
        )

    def _fit_core(self, s: pd.Series) -> None:
        self._sync_params()
        self.pipe_.fit(s)
        self.freq_ = self.pipe_.steps["deseasonal_residual"]["model"].freq_
        self.seasonal_ = self.pipe_.steps["deseasonal_residual"][
            "model"
        ].seasonal_

    def _predict_core(self, s: pd.Series) -> pd.Series:
        self._sync_params()
        return self.pipe_.detect(s)



"""Module for aggregators.

An aggregator combines multiple lists of anomalies into one.

"""

from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pandas as pd

from .._aggregator_base import _Aggregator
from ..data import validate_events


class CustomizedAggregator(_Aggregator):
    """Aggregator derived from a user-given function and parameters.

    Parameters
    ----------
    aggregate_func: function
        A function aggregating multiple types of anomaly.

        The first input argument must be a pandas DataFrame, a dict of pandas
        Series/DataFrame, or a dict of event lists.

        - If a pandas DataFrame, every column is a binary Series representing a
          type of anomaly.
        - If a dict of pandas Series/DataFrame, every value of the dict is a
          binary Series/DataFrame representing a type or some types of anomaly;
        - If a dict of list, every value of the dict is a type of anomaly as a
          list of events, where each event is represented as a pandas Timestamp
          if it is instantaneous or a 2-tuple of pandas Timestamps if it is a
          closed time interval.

        Optional input argument may be accepted through parameter
        `aggregate_func_params`.

        The output must be a list of pandas Timestamps.

        - If input is a pandas DataFrame or a dict of Series/DataFrame, return
          a single binary pandas Series;
        - If input is a dict of lists, return a single list of events.

    aggregate_func_params: dict, optional
        Parameters of `aggregate_func`. Default: None.

    """

    def __init__(
        self,
        aggregate_func: Callable,
        aggregate_func_params: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__()
        self.aggregate_func = aggregate_func
        self.aggregate_func_params = aggregate_func_params

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return ("aggregate_func", "aggregate_func_params")

    def _predict_core(
        self,
        lists: Union[
            pd.DataFrame,
            Dict[str, Union[pd.Series, pd.DataFrame]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
    ) -> Union[
        pd.Series, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ]:
        if self.aggregate_func_params is None:
            aggregate_func_params = {}
        else:
            aggregate_func_params = self.aggregate_func_params
        return self.aggregate_func(lists, **aggregate_func_params)


class OrAggregator(_Aggregator):
    """Aggregator that identifies a time point as anomalous as long as it is
    included in one of the input anomaly lists.
    """

    def __init__(self) -> None:
        super().__init__()

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return tuple()

    def _predict_core(
        self,
        lists: Union[
            pd.DataFrame,
            Dict[str, Union[pd.Series, pd.DataFrame]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
    ) -> Union[
        pd.Series, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ]:
        if isinstance(lists, dict):
            if isinstance(next(iter(lists.values())), list):
                clean_lists = {
                    key: validate_events(value) for key, value in lists.items()
                }
                return validate_events(
                    [
                        window
                        for clean_predict in clean_lists.values()
                        for window in clean_predict
                    ]
                )
            else:  # a dict of pandas Series/DataFrame
                return self._predict_core(
                    pd.concat(lists, join="outer", axis=1)
                )
        else:  # pandas DataFrame
            predicted = lists.any(axis=1)
            predicted[~predicted & lists.isna().any(axis=1)] = float("nan")
            return predicted


class AndAggregator(_Aggregator):
    """Aggregator that identifies a time point as anomalous only if it is
    included in all the input anomaly lists.
    """

    def __init__(self) -> None:
        super().__init__()

    @property
    def _param_names(self) -> Tuple[str, ...]:
        return tuple()

    def _predict_core(
        self,
        lists: Union[
            pd.DataFrame,
            Dict[str, Union[pd.Series, pd.DataFrame]],
            Dict[
                str,
                List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]],
            ],
        ],
    ) -> Union[
        pd.Series, List[Union[Tuple[pd.Timestamp, pd.Timestamp], pd.Timestamp]]
    ]:
        if isinstance(lists, dict):
            if isinstance(next(iter(lists.values())), list):
                clean_lists = {
                    key: validate_events(value, point_as_interval=True)
                    for key, value in lists.items()
                }
                time_window_stats = {
                    key: pd.Series(
                        [0] * len(clean_predict)
                        + [1] * 2 * len(clean_predict)
                        + [0] * len(clean_predict),
                        index=(
                            [
                                window[0] - pd.Timedelta("1ns")
                                for window in clean_predict
                            ]
                            + [window[0] for window in clean_predict]
                            + [window[1] for window in clean_predict]
                            + [
                                window[1] + pd.Timedelta("1ns")
                                for window in clean_predict
                            ]
                        ),
                        dtype=int,
                    ).sort_index()
                    for key, clean_predict in clean_lists.items()
                }  # type: Union[Dict, pd.Series]
                time_window_stats = {
                    key: value[~value.index.duplicated()]
                    for key, value in time_window_stats.items()
                }
                time_window_stats = (
                    pd.concat(time_window_stats, axis=1, join="outer")
                    .fillna(method="ffill")
                    .fillna(method="bfill")
                    .fillna(0)
                )
                time_window_stats = time_window_stats.all(axis=1)
                status = 0
                last_t = None
                aggregated_predict = []
                for t, v in time_window_stats.items():
                    if (status == 0) and (v == 1):
                        start = t
                        status = 1
                    if (status == 1) and (v == 0):
                        end = last_t
                        aggregated_predict.append((start, end))
                        status = 0
                    last_t = t
                return validate_events(aggregated_predict)
            else:  # a dict of pandas Series/DataFrame
                return self._predict_core(
                    pd.concat(lists, join="outer", axis=1)
                )
        else:  # pandas DataFrame
            predicted = lists.all(axis=1)
            predicted[predicted & lists.isna().any(axis=1)] = float("nan")
            return predicted
